/*
 * This file is part of the TDE Character Generator (TCG).
 *
 * TCG is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * TCG is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with TCG.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.bernhardjung.dsaprogramm.bausteine.bonus;

import static de.bernhardjung.dsaprogramm.daten.xml.HeldenbausteinXmlTags.ATTR_BONUS;
import static de.bernhardjung.dsaprogramm.daten.xml.HeldenbausteinXmlTags.ATTR_HAUSZAUBER;
import static de.bernhardjung.dsaprogramm.daten.xml.HeldenbausteinXmlTags.ATTR_VARIANTE;
import static de.bernhardjung.dsaprogramm.daten.xml.HeldenbausteinXmlTags.ATTR_ZAUBER;
import static de.bernhardjung.dsaprogramm.daten.xml.HeldenbausteinXmlTags.TAG_ZAUBER_BONUS;
import net.n3.nanoxml.IXMLElement;
import net.n3.nanoxml.XMLElement;
import de.bernhardjung.dsaprogramm.basis.SKT;
import de.bernhardjung.dsaprogramm.basis.varianten.Variante;
import de.bernhardjung.dsaprogramm.i18n.I18n;
import de.bernhardjung.dsaprogramm.zauber.Zauber;

public class Zauberbonus {

    /** Gibt an ob es sich um einen Hauszauber handelt. */
    private final boolean hauszauber;

    /** Der Stufenbonus den der Held erhaelt. */
    private final int bonus;

    /** Die moegliche Variante fuer den Zauber. */
    private final Variante variante;

    /**
     * Die Id, die in der XML-Datei fuer die Variante verwendet wurde,
     * oder einfach der Text der Variante.
     */
    private final String variantenId;

    /** Der Zauber den man durch den Bonus erhaelt. */
    private final Zauber zauber;

    /**
     * @param derZauber
     *            den man durch den Bonus erhaelt.
     * @param derStufenbonus
     * @param istHauszauber
     *            {@code true} wenn es sich um einen Hauszauber handelt, sonst
     *            {@code false}.
     */
    public Zauberbonus(
            final Zauber derZauber,
            final int derStufenbonus,
            final boolean istHauszauber ) {

        this ( derZauber, null, derStufenbonus, istHauszauber );
    }

    /**
     * @param derZauber
     *            den man durch den Bonus erhaelt.
     * @param derStufenbonus
     *            den der Held auf den Zauber erhaelt.
     * @param dieVariantenId
     *            Die Id die in der XML-Datei fuer die Variante verwendet wird
     *            oder der Text fuer die Variante.
     * @param istHauszauber
     *            {@code true} wenn es sich um einen Hauszauber handelt, sonst
     *            {@code false}.
     */
    public Zauberbonus(
            final Zauber derZauber,
            final String dieVariantenId,
            final int derStufenbonus,
            final boolean istHauszauber ) {

        hauszauber = istHauszauber;

        bonus = derStufenbonus;

        variantenId = dieVariantenId;

        if ( null != variantenId ) {

            variante = new Variante( variantenId );

        } else {

            variante = null;
        }

        zauber = derZauber;
    }

    /**
     * @return den Stufenbonus den ein Held auf den Zauber erhaelt.
     */
    public final int getBonus() {
        return bonus;
    }

    /**
     * @return die GP-Kosten fuer einen Heldenbaustein, die durch diesen Bonus
     *         entstehen.
     */
    public final double getGpKosten() {

        int spalte = zauber.getSteigerungsspalte();

        if ( hauszauber ) {

            --spalte;
        }

        final int faktor = SKT.getSteigerungsfaktor( spalte );

        final double kosten = ( bonus * faktor ) + faktor;

        return kosten / 10.0;
    }

    /** @return die moegliche Variante fuer den Zauber. */
    public final Variante getVariante() {
        return variante;
    }

    /** @return den Zauber den man durch den Bonus erhaelt. */
    public final Zauber getZauber() {
        return zauber;
    }

    /** @return {@code true}, wenn es sich um einen Hauszauber handelt, sonst
     *          {@code false}.
     */
    public final boolean istHauszauber() {
        return hauszauber;
    }

    /** @return die String-Repraesentation des Bonus. */
    @Override
    public String toString() {

        final StringBuilder string = new StringBuilder();

        string.append( zauber.getName() );

        if ( null != variante ) {

            string.append( " (" );
            string.append( variante );
            string.append( ")" );
        }

        string.append( " +" );
        string.append( bonus );

        if ( hauszauber ) {

            string.append( " " );
            string.append( I18n.i18n( "ZauberBonus.Hauszauber" ) );
        }

        return string.toString();
    }

    public final IXMLElement toXml() {

        final IXMLElement element = new XMLElement( TAG_ZAUBER_BONUS );

        element.setAttribute( ATTR_ZAUBER, zauber.getId() );

        element.setAttribute( ATTR_BONUS, String.valueOf( bonus ) );

        element.setAttribute( ATTR_HAUSZAUBER, String.valueOf( hauszauber ) );

        if ( null != variantenId ) {

            element.setAttribute( ATTR_VARIANTE, variantenId );
        }

        return element;
    }
}
