/*
 * This file is part of the TDE Character Generator (TCG).
 *
 * TCG is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * TCG is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with TCG.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.bernhardjung.dsaprogramm.erschaffung;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Observable;
import java.util.Queue;

import de.bernhardjung.dsaprogramm.basis.GeldBoerse;
import de.bernhardjung.dsaprogramm.basis.Held;
import de.bernhardjung.dsaprogramm.basis.Lernmethode;
import de.bernhardjung.dsaprogramm.bausteine.Heldenbaustein;
import de.bernhardjung.dsaprogramm.bausteine.Kultur;
import de.bernhardjung.dsaprogramm.bausteine.Profession;
import de.bernhardjung.dsaprogramm.bausteine.Rasse;
import de.bernhardjung.dsaprogramm.bausteine.bonus.SonderfertigkeitBonus;
import de.bernhardjung.dsaprogramm.bausteine.bonus.Talentbonus;
import de.bernhardjung.dsaprogramm.bausteine.bonus.TalentbonusFest;
import de.bernhardjung.dsaprogramm.bausteine.bonus.TalentbonusTyp;
import de.bernhardjung.dsaprogramm.bausteine.bonus.VerbilligteSonderfertigkeit;
import de.bernhardjung.dsaprogramm.bausteine.bonus.VerbilligteVariante;
import de.bernhardjung.dsaprogramm.bausteine.bonus.VorNachteilBonus;
import de.bernhardjung.dsaprogramm.bausteine.bonus.Zauberbonus;
import de.bernhardjung.dsaprogramm.eigenschaften.Eigenschaft;
import de.bernhardjung.dsaprogramm.gui.steigern.PanelSteigern;
import de.bernhardjung.dsaprogramm.sonderfertigkeiten.Sonderfertigkeit;
import de.bernhardjung.dsaprogramm.sonderfertigkeiten.SonderfertigkeitWert;
import de.bernhardjung.dsaprogramm.sonderfertigkeiten.Sonderfertigkeitgruppe;
import de.bernhardjung.dsaprogramm.talente.Sprache;
import de.bernhardjung.dsaprogramm.talente.Talent;
import de.bernhardjung.dsaprogramm.talente.TalentWert;
import de.bernhardjung.dsaprogramm.talente.Talentgruppe;
import de.bernhardjung.dsaprogramm.util.StringL10n;
import de.bernhardjung.dsaprogramm.vornachteile.VorNachteil;
import de.bernhardjung.dsaprogramm.vornachteile.VorNachteilWert;
import de.bernhardjung.dsaprogramm.zauber.Zauber;
import de.bernhardjung.dsaprogramm.zauber.ZauberWert;

public class ErschaffungManager extends Observable {

    /** Die AP die ein Held pro Punkt im Vorteil Gebildet erhaelt. */
    private static final int AP_PRO_STUFE_GEBILDET = 40;

    private static final String LEITZAUBER = "ErschaffungManager.Leitzauber";
    
    /** Die GP mit denen ein Held erschaffen wird. */
    private static final int START_GP = 110;

    /** Die Anzahl zusaetzlicher Leitzauber, die der Held waehlen darf. */
    private int anzahlZusaetzlicheLeitzauber;

    /** Der Held der gerade erschaffen wird. */
    private final Held held;

    /**
     * Die Liste der Boni auf Sonderfertigkeiten bei denen der Benutzer eine
     * Auswahl treffen muss.
     */
    private final Queue< SonderfertigkeitBonus > sonderfertigkeitBoniOffen;

    /**
     * Hier wird gespeicher wie bei der Erschaffung die einzelnen
     * Sonderfertigkeiten bezahlt werden. Sonderfertigkeiten koennen mit AP oder
     * GP bezahlt werden.
     */
    private final Map< SonderfertigkeitWert, Zahlungsart > sonderfertigkeitenZahlungsart;

    /**
     * Die Liste der Talent-Boni die es dem Benutzer erlauben zu waehlen.
     * Enthaelt die Boni aus Rasse und Kultur. Die Boni aus der Profession
     * befinden sich in {@link #talentBoniOffenProfession}.
     */
    private final Queue< Talentbonus > talentBoniOffen;

    /**
     * Die Talent-Boni der Profession die es erlauben eine Auswahl zu treffen.
     * Die restlichen Boni befinden sich in {@link #talentBoniOffen}
     */
    private final Queue< Talentbonus > talentBoniOffenProfession;

    /**
     * Ein Spezieller Talentbonus der nach allen anderen abgehandelt werden
     * muss. Mehr dazu siehe Regelwerk Seite 76, Profession Entdecker, Kategorie
     * Wissen.
     */
    private TalentbonusFest talentBonusEntdecker;

    /**
     * Hier wird eine Bonus gespeichert bei welchem der Benutzer aus eine Liste
     * eine Sonderfertigkeit auswaehlen darf, die dann fuer den Helden
     * verbilligt zu erwerben ist. Das kommt im ganzen Regelwerk nur einmal vor,
     * daher braucht es hier keine Queue.
     */
    private VerbilligteSonderfertigkeit verbilligteSonderfertigkeit;

    /**
     * Die Punkte die der Held durch Rasse, Kultur und Profession erhalten hat
     * um sie auf den Vorteil Verbindungen zu verteilen.
     */
    private int verbindungen;

    /**
     * Die Boni auf die Vor-/Nachteile, die es dem Benutzer erlauben einen
     * auszuwaehlen.
     */
    private final Queue< VorNachteilBonus > vorNachteilBoniOffen;

    /**
     * Die Zweitsprache die die Kultur des Helden zur Verfuegung stellt.
     */
    private final List< Sprache > zweitsprachen;

    public ErschaffungManager() {

        held = new Held();

        sonderfertigkeitBoniOffen = new LinkedList< SonderfertigkeitBonus >();

        sonderfertigkeitenZahlungsart = new HashMap< SonderfertigkeitWert, Zahlungsart >();

        talentBoniOffen = new LinkedList< Talentbonus >();

        talentBoniOffenProfession = new LinkedList< Talentbonus >();

        verbindungen = 0;

        vorNachteilBoniOffen = new LinkedList< VorNachteilBonus >();

        zweitsprachen = new ArrayList< Sprache >( 2 );
    }

    /**
     * Wendet einen Talent-Bonus auf den Helden an, wenn fuer den Talent-Bonus
     * eine Auswahl getroffen werden muss wird er in {@link #talentBoniOffen}
     * gespeichert.
     *
     * @param bonus
     *            der auf den Helden angewendet werden soll.
     */
    private void anwenden( final TalentbonusFest bonus ) {

        if (bonus.istDirektAnwendbar( held )) {

            for ( final TalentWert wert : bonus.getTalentboni( held ) ) {

                held.einfuegenVorgegeben( wert );

                if (bonus.istLeittalent()) {

                    held.einfuegenLeittalent( wert.getTalent() );

                }

            }

        } else {

            talentBoniOffen.add( bonus );
        }

    }

    /**
     * Wendet alle Sonderfertigkeiten und verbilligte Sonderfertigkeiten auf den
     * Helden an. Die Boni fuer die vom Benutzer noch eine Auswahl getroffen
     * werden muss, werden gespeichert und koennen ueber
     * {@link #getSonderfertigkeitBonusOffen()} und
     * {@link #getVerbilligteSonderfertigkeitOffen()} abgerufen werden.
     */
    final void anwendenSonderfertigkeiten() {

        // anwendenSonderfertigkeiten( held.getRasse() );

        final Kultur kultur = held.getKultur();

        anwendenSonderfertigkeiten( kultur );

        anwendenSonderfertigkeitenVerbilligt( kultur );

        held.einfuegenLeitsonderfertigkeiten( kultur
                .getLeitsonderfertigkeiten() );

        final Profession profession = held.getProfession();

        anwendenSonderfertigkeiten( profession );

        anwendenSonderfertigkeitenVerbilligt( profession );
    }

    /**
     * Wendet alle Sonderfertigkeiten die der Held durch einen Baustein erhaelt
     * auf ihn an. Boni fuer die noch eine Auswahl getroffen werden muss werden
     * gespeichert und koennen ueber {@link #getSonderfertigkeitBonusOffen()}
     * abgerufen werden.
     *
     * @param baustein
     *            dessen Sonderfertigkeit-Boni auf den Helden angewendet werden
     *            sollen.
     */
    private void anwendenSonderfertigkeiten( final Heldenbaustein baustein ) {

        final List< SonderfertigkeitBonus > boni;
        boni = baustein.getSonderfertigkeiten();

        for ( final SonderfertigkeitBonus bonus : boni ) {

            final List< SonderfertigkeitWert > werte = bonus.getWerte( held );

            if (werte.size() > 1) {

                sonderfertigkeitBoniOffen.add( bonus );

            } else {

                // Es gibt nichts zu waehlen, da die Liste werte nur einen
                // Eintrag enthaelt. Diesen kann der Held direkt erhalten.
                einfuegenVorgegeben( werte.get( 0 ) );

            }
        }
    }

    /**
     * Wendet alle Verbilligungen die ein Baustein vergibt auf den Helden an.
     * Wenn es noch eine Verbilligung gibt bei der aus einer Liste eine
     * ausgewaehlt werden darf wird diese in {@link VerbilligteSonderfertigkeit}
     * gespeichert.
     *
     * @param baustein
     *            dessen Verbilligungen auf den Helden angewendet werden sollen.
     */
    private void anwendenSonderfertigkeitenVerbilligt( final Heldenbaustein baustein ) {

        final List< VerbilligteSonderfertigkeit > verbilligungen;
        verbilligungen = baustein.getVerbilligteSonderfertigkeiten();

        for ( final VerbilligteSonderfertigkeit verbilligung : verbilligungen ) {

            if (verbilligung.hatAuswahl()) {

                verbilligteSonderfertigkeit = verbilligung;

            } else {

                einfuegenVerbilligt( verbilligung.getSonderfertigkeit() );

            }
        }

        final List< VerbilligteVariante > verbilligteVarianten;
        verbilligteVarianten = baustein.getVerbilligteVarianten();

        for ( final VerbilligteVariante variante : verbilligteVarianten ) {

            held.einfuegenVerbilligt( variante.getWert() );

        }
    }

    /**
     * Nachdem fuer einen Helden Rasse, Kultur und Profession gewaehlt wurden,
     * koennen dem Held durch Aufruf dieser Methode die Talent-Boni
     * gutgeschrieben werden. Talent-Boni fuer die noch eine Auswahl getroffen
     * werden muss werden gespeichert und koennen ueber die Methoden
     * {@link #getTalentBonusOffen}, {@link #getTalentBonusOffenProfession()}
     * und {@link #getTalentBonusEntdecker()} abgerufen werden.
     * <p>
     * Die Methode uebernimmt auch die Leittalente von Rasse und Kultur.
     *
     * @see #getTalentBonusOffen()
     * @see #getTalentBonusOffenProfession()
     * @see #hatTalentBonusOffen()
     * @see #hatTalentBonusOffenProfession()
     * @see #hatTalentBonusEntdecker()
     * @see #getTalentBonusEntdecker()
     */
    final void anwendenTalentBoniLeittalente() {

        held.einfuegenLeittalente( held.getRasse().getLeittalente() );

        held.einfuegenLeittalente( held.getKultur().getLeittalente() );

        anwendenTalentBoniRasseKultur();

        anwendenTalentBoniProfession();
    }

    /**
     * Wendet die Talent-Boni eines Heldenbausteins auf den Helden an.
     *
     * @param baustein
     *            dessen Talent-Boni auf den Helden angewendet werden sollen.
     */
    private void anwendenTalentBoni( final Heldenbaustein baustein ) {

        for ( final Talentgruppe gruppe : Talentgruppe.getTalentgruppen() ) {

            final List< Talentbonus > boni = baustein.getTalentboni( gruppe );

            for ( final Talentbonus bonus : boni ) {

                switch (bonus.getTyp()) {

                case FEST:
                    anwenden( (TalentbonusFest) bonus );
                    break;

                default:
                    talentBoniOffen.add( bonus );
                    break;
                }
            }
        }
    }

    /**
     * Wendet die Talent-Boni, die der Held durch seine Profession erhaelt auf
     * ihn an.
     */
    private void anwendenTalentBoniProfession() {

        final Heldenbaustein baustein = held.getProfession();

        for ( final Talentgruppe gruppe : Talentgruppe.getTalentgruppenAlle() ) {

            final List< Talentbonus > boni = baustein.getTalentboni( gruppe );

            for ( final Talentbonus bonus : boni ) {

                switch (bonus.getTyp()) {

                case FEST:
                    final TalentbonusFest bonusFest = (TalentbonusFest) bonus;

                    if (bonusFest.istDirektAnwendbar( held )) {

                        final List< TalentWert > werte;
                        werte = bonusFest.getTalentboni( held );

                        for ( final TalentWert wert : werte ) {

                            held.einfuegenProfession( wert );

                            if (bonusFest.istLeittalent()) {

                                held.einfuegenLeittalent( wert.getTalent() );

                            }

                        }

                    } else {

                        talentBoniOffenProfession.add( bonus );
                    }

                    break;

                case FREI:
                    talentBoniOffenProfession.add( bonus );
                    break;

                case ENTDECKER:
                    talentBonusEntdecker = (TalentbonusFest) bonus;
                    break;

                }
            }
        }
    }

    /**
     * Wendet die Talent-Boni die der Held durch Rasse und Kultur erhaelt auf
     * ihn an.
     */
    private void anwendenTalentBoniRasseKultur() {

        anwendenTalentBoni( held.getRasse() );

        final Kultur kultur = held.getKultur();

        held.setMuttersprache( kultur.getMuttersprache() );

        zweitsprachen.addAll( kultur.getZweitsprachen() );

        if (zweitsprachen.size() == 1) {
            held.setZweitsprache( zweitsprachen.get( 0 ) );
        }

        anwendenTalentBoni( held.getKultur() );
    }

    /**
     * Wendet alle Vor-/Nachteile auf den Helden an. Vor-/Nachteile fuer die
     * noch eine Auswahl offen ist koennen im Anschluss an diesen Methodenaufruf
     * hier abgerufen werden.
     *
     * @see #getVorNachteilBonusOffen()
     */
    final void anwendenVorNachteile() {

        final Rasse rasse = held.getRasse();
        anwendenVorNachteile( rasse.getVorteile() );
        anwendenVorNachteile( rasse.getNachteile() );

        final Kultur kultur = held.getKultur();
        anwendenVorNachteile( kultur.getVorteile() );
        anwendenVorNachteile( kultur.getNachteile() );

        final Profession profession = held.getProfession();
        anwendenVorNachteile( profession.getVorteile() );
        anwendenVorNachteile( profession.getNachteile() );

        verbindungen = rasse.getVerbindungen()
                     + kultur.getVerbindungen()
                     + profession.getVerbindungen();
    }

    /**
     * Wendet eine Liste von Vor-/Nachteilen auf den Helden an. Speichert die
     * Boni fuer die eine Auswahl getroffen werden muss in
     * {@link #vorNachteilBoniOffen},
     *
     * @param boni
     *            die angewendet werden sollen.
     */
    private void anwendenVorNachteile( final List< VorNachteilBonus > boni ) {

        for ( final VorNachteilBonus bonus : boni ) {

            final List< VorNachteilWert > werte = bonus.getWerte();

            if (1 == werte.size()) {

                einfuegenVorgegeben( werte.get( 0 ) );

            } else {

                vorNachteilBoniOffen.add( bonus );

            }
        }
    }

    /** Wendet die Zauberboni auf den Helden an. */
    final void anwendenZauber() {

        final Kultur kultur = held.getKultur();

        anzahlZusaetzlicheLeitzauber = kultur.getAnzahlLeitzauber();

        anwendenZauber( kultur.getZauberboni() );

        anwendenZauber( held.getProfession().getZauberboni() );
    }

    /**
     * Wendet eine Liste von Zaubern auf den Helden an.
     *
     * @param boni
     *            die der Held erhalten soll.
     */
    private void anwendenZauber( final List< Zauberbonus > boni ) {

        for ( final Zauberbonus bonus : boni ) {

            final Zauber zauber = bonus.getZauber();

            final ZauberWert wert;

            if (zauber.hatVarianten()) {

                wert = new ZauberWert( 
                        zauber, 
                        bonus.getVariante(), 
                        bonus.getBonus() );

            } else {

                wert = new ZauberWert( zauber, bonus.getBonus() );

            }

            held.einfuegenVorgegeben( wert );

            if (held.hat( VorNachteil.ELFISCHE_WELTSICHT )) {

                einfuegenLeitzauber( zauber );
            }

            if (bonus.istHauszauber()) {

                held.einfuegenHauszauber( zauber );
            }
        }
    }

    /**
     * Markiert eine Sonderfertigkeit als fuer den Helden verbilligt. Er kann
     * sie zu halben Kosten erlernen.
     *
     * @param sonderfertigkeit
     *            die verbilligt wird.
     */
    final void einfuegenVerbilligt( final Sonderfertigkeit sonderfertigkeit ) {

        if (!sonderfertigkeit.hatVarianten()) {
            // Eine doppelt verbilligten Sonderfertigkeit wird automatisch
            // aktiviert.
            final SonderfertigkeitWert wert;
            wert = new SonderfertigkeitWert( sonderfertigkeit );

            if (held.hatVerbilligt( wert )) {

                einfuegenVorgegeben( wert );
            }
        }

        held.einfuegenVerbilligt( sonderfertigkeit );
    }

    /**
     * Fuegt dem Helden der gerade erschaffen wird eine Sonderfertigkeit als
     * durch Rasse, Kultur oder Profession vorgegeben hinzu.
     *
     * @param wert
     *            den der Held erhalten soll.
     */
    final void einfuegenVorgegeben( SonderfertigkeitWert wert ) {

        final Sonderfertigkeit sonderfertigkeit;
        sonderfertigkeit = wert.getSonderfertigkeit();

        // Erhaelt der Held 2-mal Ruestungsgewoehnung I erhält er
        // Ruestungsgewoehnung II
        if (( Sonderfertigkeit.RUESTUNGSGEWOEHNUNG_I == sonderfertigkeit )
                && held.hat( Sonderfertigkeit.RUESTUNGSGEWOEHNUNG_I )) {

            wert = new SonderfertigkeitWert(
                    Sonderfertigkeit.RUESTUNGSGEWOEHUNG_II );

        }

        held.einfuegenVorgegeben( wert );
    }

    /**
     * Fuegt dem Helden der gerade erschaffen wird einen Talent-Bonus, als durch
     * Rasse oder Kultur vorgegeben hinzu.
     *
     * @see #einfuegenProfession(TalentWert)
     * @param wert
     *            der dem Helden durch Rasse oder Kultur vorgegeben wird.
     */
    final void einfuegenVorgegeben( final TalentWert wert ) {
        held.einfuegenVorgegeben( wert );
    }

    /**
     * Fuegt dem Helden der gerade erschaffen wird einen Vor-/Nachteil als durch
     * Kultur oder Profession vorgegeben hinzu.
     *
     * @param wert
     *            den der Held erhalten soll.
     */
    final void einfuegenVorgegeben( final VorNachteilWert wert ) {

        if (wert.getVorNachteil().equals( VorNachteil.VOLLZAUBERER )
                && held.hat( VorNachteil.VIERTELZAUBERER )) {

            held.entfernenVorgegeben( VorNachteil.VIERTELZAUBERER );

        }

        if (!wert.getVorNachteil().equals( VorNachteil.VIERTELZAUBERER )
                || !held.getProfession().istKaempferprofession()) {

            held.einfuegenVorgegeben( wert );
        }

    }

    /**
     * Markiert einen Zauber als Leitzauber fuer den Helden.
     *
     * @param zauber
     *            der als Leitzauber markiert werden soll.
     */
    final void einfuegenLeitzauber( final Zauber zauber ) {
        held.einfuegenLeitzauber( zauber );
    }

    /**
     * @see #einfuegenVorgegeben(TalentWert)
     * @param wert
     *            der dem Helden durch seine Profession vorgegeben wird.
     */
    final void einfuegenProfession( final TalentWert wert ) {
        held.einfuegenProfession( wert );
    }

    /**
     * Stellt die Eigenschaftsstartstufen des Helden auf die geforderten
     * Mindeststufe.
     */
    final void einstellenMindeststufe() {

        for ( final Eigenschaft eigenschaft : Eigenschaft
                .getEigenschaftenMitSo() ) {

            // Die Startstufe darf auf jeden Fall nicht unter diesem Wert
            // liegen.
            int mindeststufe = eigenschaft.getMinBasisstufe( held );

            final Rasse rasse = held.getRasse();

            final int modifikation = eigenschaft.getModifikation( held );

            if (rasse.hatMindeststufe( held, eigenschaft )) {
                // Die Mindeststufe durch Rasse ist auf jeden Fall
                // groesser als die MinStartstufe, daher
                // kann man sie einfach uebernehmen.
                mindeststufe = rasse.getMindeststufe( eigenschaft );

                // Allerdings muss dieser Wert erst nach addieren der
                // modifikation erreicht sein. Daher muss man die
                // modifikation noch abziehen. Auch jetzt ist die
                // mindeststufe noch hoeher als die MinStartstufe
                mindeststufe -= modifikation;

            }

            final Profession profession = held.getProfession();

            if (profession.hatMindeststufe( held, eigenschaft )) {

                int mindeststufeProfession = profession
                        .getMindeststufe( eigenschaft );

                // sieh oben.
                mindeststufeProfession -= modifikation;

                mindeststufe = Math.max( mindeststufe, mindeststufeProfession );
            }

            held.getWert( eigenschaft ).setBasisstufe( mindeststufe );
        }
    }

    /**
     * @return Die Anzahl zusaetzlicher Leitzauber, die der Held waehlen darf.
     */
    int getAnzahlZusaetzlicheLeitzauber() {
        return anzahlZusaetzlicheLeitzauber;
    }

    /**
     * @return die AP die dem Helden, durch KL und IN, zur Erschaffung zur
     *         Verfuegung stehen. Das beinhaltet nicht die AP durch Gebildet.
     */
    private int getAp() {

        final int kl = held.getStufe( Eigenschaft.KL );

        final int in = held.getStufe( Eigenschaft.IN );

        return 20 * ( kl + in );
    }

    /**
     * Berechnet wie viele AP durch den Vorteil Gebildet eingespart werden bzw.
     * wie viele der AP die der Charakter durch Gebildet erhalten hat verbraucht
     * werden.
     *
     * @return die durch Gebildet eingesparten GP oder 0 wenn der Charakter den
     *         Vorteil Gebildet nicht besitzt.
     */
    private int getApEinsparungGebildet() {

        int einsparung = 0;

        if ( held.hat( VorNachteil.GEBILDET ) ) {

            // Die AP-Kosten von Talenten und nichtmagischen Sonderfertigkeiten.
            int apKosten = 0;

            apKosten += getApKostenTalente(
                    held.getTalente( Talentgruppe.WISSEN ) );

            apKosten += getApKostenTalente(
                    held.getTalente( Talentgruppe.SPRACHEN ) );

            apKosten += getApKostenTalente(
                    held.getTalente( Talentgruppe.HANDWERK ) );

            for ( final SonderfertigkeitWert wert :
                                        held.getSonderfertigkeiten() ) {

                if ( ! wirdMitGpBezahlt( wert ) ) {

                    final Sonderfertigkeit sf = wert.getSonderfertigkeit();

                    // Obacht! talent kann null sein.
                    final Talent talent = sf.getTalent();

                    if ( sf.istTalentspezialisierung()
                            && ( talent.getTalentgruppe().equals(
                                    Talentgruppe.WISSEN )
                              || talent.getTalentgruppe().equals(
                                    Talentgruppe.SPRACHEN )
                              || talent.getTalentgruppe().equals(
                                    Talentgruppe.HANDWERK ) ) ) {

                        apKosten += wert.getApKosten( held );

                    } else if ( sf
                            .equals( Sonderfertigkeit.NANDUSGEFAELLIGES_WISSEN )) {

                        apKosten += wert.getApKosten( held );

                    }
                }
            }

            final int stufe = held.getWert( VorNachteil.GEBILDET ).getStufe();

            final int apGebildet = stufe * AP_PRO_STUFE_GEBILDET;

            // Die Haelfte der apGebildet kann in magisches investiert werden.
            final int einsparungMagie = Math.min( getApKostenMagie(),
                                                  ( apGebildet + 1 ) / 2 );

            einsparung = Math.min( apGebildet, ( apKosten + einsparungMagie ) );
        }

        return einsparung;
    }

    /**
     * @return die AP, die dem Helden aus dem Vorteil Gebildet zur Verfuegung
     *         stehen. Wenn der Held den Vorteil nicht besitzt, liefert die
     *         Methode 0.
     */
    private int getApGebildet() {

        if ( held.hat( VorNachteil.GEBILDET ) ) {

            final int stufe = held.getWert( VorNachteil.GEBILDET ).getStufe();

            return ( AP_PRO_STUFE_GEBILDET * stufe );

        } else {

            return 0;
        }
    }

    /**
     * @return die Anzahl an AP die der Held bei der Erschaffung noch einsetzen
     *         kann. Dieser Wert bezieht sich nur auf die AP durch KL und IN,
     *         die AP durch Gebildet sind nicht eingerechnet.
     */
    public final int getApGuthaben() {

        int restAp = getAp();

        restAp -= getApKostenTalente( held.getTalente() );

        restAp -= getApKostenZauber( held.getZauber() );

        for ( final SonderfertigkeitWert wert : held.getSonderfertigkeiten() ) {

            if (!wirdMitGpBezahlt( wert )) {

                restAp -= wert.getApKosten( held );
            }
        }

        restAp += getApEinsparungGebildet();

        return restAp;
    }

    /**
     * Falls der Held den Vorteil Gebildet besitzt, so liefert diese Methode die
     * AP, die dem Held noch durch den Vorteil zur Verfuegung stehen.
     *
     * @return die restlichen AP aus dem Vorteil Gebildet. Die Methode liefert
     *         0, wenn der Held den Vorteil nicht besitzt.
     */
    public final int getApGuthabenGebildet() {

        int restAp = getApGebildet();

        restAp -= getApEinsparungGebildet();

        return restAp;
    }

    /**
     * @return die AP, die noch fuer Zauber und magische Sonderfertigkeiten
     *         verwendet werden koennen.
     */
    public final int getApGuthabenMagie() {

        // Die Haelfte der AP koennen in magisches investiert werden.
        int apGuthabenMagie = Math.round( getAp() / 2F );

        // dazu kommen die Haelft der AP aus Gebildet.
        apGuthabenMagie += Math.round( getApGebildet() / 2F );

        apGuthabenMagie -= getApKostenMagie();

        if ( apGuthabenMagie < 0 ) {
            // Das Guthaben ist ueberzogen. Das soll dem Benutzer angezeigt
            // werden, daher wird der Wert zurueckgegeben.
            return apGuthabenMagie;

        } else {
            // Das Guthaben ist positiv.

            // Das Guthaben kann nicht groesser sein, als das gesamte AP-Guthaben.
            apGuthabenMagie = Math.min( apGuthabenMagie, getApGuthaben() );

            // Das Guthaben kann nicht unter 0 sinken.
            return Math.max( 0, apGuthabenMagie );
        }

    }

    /**
     * @return die AP, die fuer Zauber und magische Sonderfertigkeiten bezahlt
     *         werden muessen.
     */
    private int getApKostenMagie() {

        int apKosten = getApKostenZauber( held.getZauber() );

        for ( final SonderfertigkeitWert wert :
                                    held.getSonderfertigkeiten() ) {

            if ( ! wirdMitGpBezahlt( wert ) ) {

                final Sonderfertigkeit sf = wert.getSonderfertigkeit();

                final Sonderfertigkeitgruppe gruppe = sf.getGruppe();

                if ( gruppe.equals( Sonderfertigkeitgruppe.MAGISCHE ) ) {
                  // || gruppe.equals( Sonderfertigkeitgruppe.ELFENLIEDER )
                  // || gruppe.equals( Sonderfertigkeitgruppe.STABZAUBER ) ) {

                    apKosten += wert.getApKosten( held );
                }
            }
        }

        return apKosten;
    }

    /**
     * Berechnet die AP-Kosten die ein Held fuer eine Liste von Talenten
     * bezahlen muss. Der Held muss alle uebergebenen Talente besitzen.
     *
     * @param talente
     *            deren Kosten berechnet werden sollen.
     * @return die AP-Kosten die zum Aktivieren und Steigern der Talent
     *         aufzubringen sind.
     * @throws NullPointerException
     *             wenn der Held eines der Talente nicht besitzt.
     */
    private int getApKostenTalente( final List< Talent > talente ) {

        int apKosten = 0;

        for ( final Talent talent : talente ) {

            // Wurde das Talent aktiviert? Dann sind Aktivierungskosten zu
            // zahlen.
            if (!held.hatVorgegeben( talent )) {
                // Das Talent wurde aktiviert.

                apKosten += talent.getAktivierungskosten( held );
            }

            final int start;

            start = held.getMindeststufe( talent );

            final int stufe = held.getStufe( talent );

            int i = start;

            for ( ; ( i < 0 ) && ( i < stufe ) ; ++i ) {

                apKosten += talent.getAktivierungskosten( held );
            }

            for ( ; i < stufe ; ++i ) {

                apKosten += talent.getSteigerungskosten( held, i );
            }

            if (held.hatVorgegeben( talent ) && ( 0 > start )
                    && !talent.istBasisTalent()
                    && held.getWert( talent ).istAktiviert()) {
                // Sonderfall Spezial-Talent mit negativer vorgegebener Stufe:
                // Hier muessen noch einmal die Aktivierungskosten bezahlt
                // werden.

                apKosten += talent.getAktivierungskosten( held );
            }
        }

        return apKosten;
    }

    /**
     * Berechnet die AP-Kosten die ein Held fuer eine Liste von Zaubern bezahlen
     * muss. Der Held muss alle uebergebenen Zauber besitzen.
     *
     * @param werte
     *            deren Kosten berechnet werden sollen.
     * @return die AP-Kosten die zum Aktivieren und Steigern der Zauber
     *         aufzubringen sind.
     * @throws NullPointerException
     *             wenn der Held einen der Zauber nicht besitzt.
     */
    private int getApKostenZauber( final List< ZauberWert > werte ) {

        int apKosten = 0;

        for ( final ZauberWert wert : werte ) {

            final Zauber zauber = wert.getZauber();

            final int start;

            if ( held.hatVorgegeben( wert ) ) {

                start = held.getMindeststufe( wert );

            } else {

                start = 0;
                // Der Zauber wurde aktiviert. Es sind Aktivierungskosten
                // zu zahlen.
                apKosten += zauber.getAktivierungskosten( held, wert );
            }

            final int stufe = held.getWert( wert ).getStufe();

            for ( int i = start ; i < stufe ; ++i ) {

                apKosten += zauber.getSteigerungskosten( held, wert, i,
                        Lernmethode.KEINE );
            }
        }

        return apKosten;
    }

    /**
     * @return eine fertige Version des Helden, die im {@link PanelSteigern}
     *         verwendet werde kann.
     */
    public Held getFertigenHeld() {

        final Held fertigerHeld = new Held();

        // wirdErschaffen auf false setzen muss vor dem Einfuegen von Werten
        // geschehen, da die Methoden sich dann anders verhalten.
        // Beim Einfuegen von Vor-/Nachteilen arbeiten die Anpasser anders, bzw.
        // nicht mehr.
        fertigerHeld.setWirdErschaffen( false );

        final int apGesamt = getAp();

        fertigerHeld.addiereApGesamt( apGesamt );

        final int apGebildet = getApEinsparungGebildet();

        fertigerHeld.addiereApGesamt( apGebildet );

        fertigerHeld.addiereApEingesetzt( apGesamt - getApGuthaben() );

        fertigerHeld.addiereApEingesetzt( apGebildet );

        fertigerHeld.setName( held.getName() );

        fertigerHeld.setRasse( held.getRasse() );

        fertigerHeld.setKultur( held.getKultur() );

        fertigerHeld.setProfession( held.getProfession() );

        fertigerHeld.setAlter( held.getAlter() );

        fertigerHeld.setAugenfarbe( held.getAugenfarbe() );

        fertigerHeld.setAussehen( held.getAussehen() );

        fertigerHeld.setGeburtstag( held.getGeburtstag() );

        fertigerHeld.setGeschlecht( held.getGeschlecht() );

        fertigerHeld.setGewicht( held.getGewicht() );

        fertigerHeld.setGroesse( held.getGroesse() );

        fertigerHeld.setHaarfarbe( held.getHaarfarbe() );

        fertigerHeld.setHintergrund( held.getHintergrund() );

        fertigerHeld.setStand( held.getStand() );

        fertigerHeld.setTitel( held.getTitel() );

        for ( final Eigenschaft eigenschaft : Eigenschaft
                .getEigenschaftenMitSo() ) {

            final int basisstufe =
                held.getWert( eigenschaft ).getBasisstufe();

            fertigerHeld.getWert( eigenschaft ).setBasisstufe( basisstufe );
        }

        for ( final Talent talent : held.getTalente() ) {

            fertigerHeld.einfuegenVorgegeben( held.getWert( talent ) );
        }

        fertigerHeld.einfuegenLeittalente( held.getLeittalente() );

        for ( final ZauberWert wert : held.getZauber() ) {

            fertigerHeld.einfuegenVorgegeben( wert );
        }

        fertigerHeld.einfuegenHauszauber( held.getHauszauber() );

        fertigerHeld.einfuegenLeitzauber( held.getLeitzauber() );

        for ( final SonderfertigkeitWert wert :
            held.getSonderfertigkeiten() ) {

            // Den Wert hier zu veraender ist ok. Nach Aufruf dieser Methode ist
            // die Erschaffung vorbei.
            if ( wert.getSonderfertigkeit().hatVarianten() ) {

                wert.setVarianteAenderbar( true );
            }

            fertigerHeld.einfuegenVorgegeben( wert );
        }

        fertigerHeld.einfuegenLeitsonderfertigkeiten(
                held.getLeitsonderfertigkeiten() );

        for ( final Sonderfertigkeit sonderfertigkeit :
            held.getVerbilligteSonderfertigkeiten() ) {

            fertigerHeld.einfuegenVerbilligt( sonderfertigkeit );
        }

        for ( final SonderfertigkeitWert wert :
            held.getVerbilligteVarianten() ) {

            fertigerHeld.einfuegenVerbilligt( wert );
        }

        for ( final VorNachteilWert wert : held.getVorNachteile() ) {

            if ( wert.getVorNachteil().hatVarianten() ) {
                // Den Wert hier zu veraender ist ok. Nach Aufruf dieser Methode ist
                // die Erschaffung vorbei.
                wert.setVarianteAenderbar( true );
            }

            fertigerHeld.einfuegenVorgegeben( wert );
        }

        final int kl = held.getStufe( Eigenschaft.KL );

        final Sprache muttersprache = held.getMuttersprache();

        fertigerHeld.setMuttersprache( muttersprache );

        fertigerHeld.einfuegenVorgegeben(
                new TalentWert( muttersprache, kl - 2 ) );

        if ( held.hatZweitsprache() ) {

            final Sprache zweitsprache = held.getZweitsprache();

            fertigerHeld.setZweitsprache( zweitsprache );

            fertigerHeld.einfuegenVorgegeben(
                    new TalentWert( zweitsprache, kl - 4 ) );
        }
        
        fertigerHeld.setGeldBoerse( new GeldBoerse(
                fertigerHeld.getWert( Eigenschaft.SO ).getStufe(), 
                fertigerHeld.getWert( VorNachteil.AUSRUESTUNGS_VORTEIL )));
        
        return fertigerHeld;
    }

    /**
     * @return die Anzahl der GP die der Held bei der Erschaffung noch einsetzen
     *         kann.
     */
    public final int getGpGuthaben() {

        double restGp = START_GP;

        restGp = restGp - held.getRasse().getGpKosten( held );
        restGp = restGp - held.getKultur().getGpKosten( held );
        restGp = restGp - held.getProfession().getGpKosten( held );

        for ( final Eigenschaft eigenschaft : Eigenschaft
                .getEigenschaftenMitSo() ) {

            restGp -= held.getWert( eigenschaft ).getBasisstufe();
        }

        // Die Mindeststufe welche die Profession des Helden vorgibt, muss
        // nicht mit GP bezahlt werden, sie kann wieder gutgeschrieben werden.
        final int startstufeSo = held.getWert( Eigenschaft.SO ).getBasisstufe();
        int gutschrift = held.getProfession().getMindeststufe( Eigenschaft.SO );
        if (held.hatVorgegeben( VorNachteil.AKADEMISCHE_AUSBILDUNG_KRIEGER )) {
            // Hat der Held AAK durch die Profession (Krieger) erhalten, so
            // stammt ein Punkt aus der Mindeststufe aus dem SO Bonus durch AAK.
            // Dieser Punkt wird nachtraeglich eingerechnet und darf nicht
            // gutgeschrieben werden.
            --gutschrift;
        }
        if (startstufeSo >= gutschrift) {
            restGp += gutschrift;
        } else {
            // Wenn die Startstufe des SO unter die Mindeststufe gesenkt wird
            // darf nicht die ganze Mindesstufe gutgeschrieben werden, sondern
            // nur die kleinere Startstufe des Helden.
            restGp += startstufeSo;
        }

        for ( final SonderfertigkeitWert wert : held.getSonderfertigkeiten() ) {

            if (wirdMitGpBezahlt( wert )) {

                restGp -= wert.getGpKosten( held );
            }
        }

        for ( final VorNachteilWert wert : held.getVorNachteile() ) {

            restGp -= wert.getGpKosten( held );
        }

        return (int) Math.floor( restGp );
    }

    /**
     * @return den Held der gerade erschaffen wird.
     */
    public final Held getHeld() {
        return held;
    }
    
    public final List< Object > getOffeneBoni() {
        
        final List< Object > boni = new ArrayList< Object >();
        
        boni.addAll( talentBoniOffen );
        
        boni.addAll( talentBoniOffenProfession );
        
        if ( hatTalentBonusEntdecker() ) {
            
            boni.add( talentBonusEntdecker );
        }
        
        boni.addAll( sonderfertigkeitBoniOffen );
        
        if ( hatVerbilligteSonderfertigkeitOffen() ) {
            
            boni.add( verbilligteSonderfertigkeit );
        }
        
        boni.addAll( vorNachteilBoniOffen );
        
        if ( hatZusaetzlicheLeitzauber() ) {
            
            boni.add( new StringL10n( 
                    LEITZAUBER, 
                    getAnzahlZusaetzlicheLeitzauber() ) );
        }
        
        return boni;
    }

    /**
     * Nach dem Aufruf von {@link #anwendenSonderfertigkeiten()} koennen hier
     * die Boni abgerufen werden bei denen der Benutzer eine Sonderfertigkeit
     * aus einer Liste waehlen muss. Die Methode liefert bei jedem Aufruf
     * nacheinander die noch offenen Auswahlen.
     *
     * @see #hatSonderfertigkeitBonusOffen()
     * @return einen Bonus fuer den noch eine Auswahl zu treffen ist, oder
     *         {@code null}, wenn es keinen offenen Bonus mehr gibt.
     */
    final SonderfertigkeitBonus getSonderfertigkeitBonusOffen() {
        return sonderfertigkeitBoniOffen.poll();
    }

    /**
     * Nach Aufruf von {@link #anwendenTalentBoniLeittalente()} kann hier ein
     * Talent-Bonus vom {@link TalentbonusTyp#ENTDECKER} abgerufen werden, so es
     * einen gibt. Dieser Talent-Bonus muss nach allen anderen abgehandelt
     * werden.
     *
     * @see #getTalentBonusOffen()
     * @see #getTalentBonusOffenProfession()
     * @see #hatTalentBonusEntdecker()
     * @return einen Talent-Bonus oder {@code null} falls es keinen offenen
     *         Talent-Bonus (mehr) gibt.
     */
    final TalentbonusFest getTalentBonusEntdecker() {

        final TalentbonusFest bonus = talentBonusEntdecker;

        talentBonusEntdecker = null;

        return bonus;
    }

    /**
     * Nach dem Aufruf von {@link #anwendenTalentBoniLeittalente()} koennen hier
     * die Talent-Boni abgerufen werden fuer die noch eine Auswahl zu treffen
     * ist. Die Methode liefert bei jedem Aufruf nacheinander die noch offenen
     * Talent-Boni. Es werden die Talent-Boni geliefert die der Held durch Rasse
     * oder Kultur erhalten hat.
     *
     * @see #hatTalentBonusOffen()
     * @return einen Talent-Bonus fuer den noch eine Auswahl zu treffen ist oder
     *         {@code null} wenn es keinen offenen Talent-Bonus mehr gibt.
     */
    final Talentbonus getTalentBonusOffen() {
        return talentBoniOffen.poll();
    }

    /**
     * Nach dem Aufruf von {@link #anwendenTalentBoniLeittalente()} koennen hier
     * die Talent-Boni abgerufen werden fuer die noch eine Auswahl zu treffen
     * ist. Die Methode liefert bei jedem Aufruf den jeweils naechsten
     * Talent-Bonus in der Liste. Es werden die Talent-Boni geliefert die der
     * Held durch seine Profession erhalten hat.
     *
     * @see #hatTalentBonusOffenProfession()
     * @return den naechsten Talent-Bonus fuer den noch eine Auswahl zu treffen
     *         ist oder {@code null} wenn es keinen offenen Talent-Bonus mehr
     *         gibt.
     */
    final Talentbonus getTalentBonusOffenProfession() {
        return talentBoniOffenProfession.poll();
    }

    /**
     * Wenn es nach dem Anwenden der verbilligten Sonderfertigkeiten noch eine
     * Bonus gibt bei dem aus einer Liste eine Sonderfertigkeit gewaehlt werden
     * muss die verbilligt wird so kann diese hier abgerufen werden.
     *
     * @return die Liste mit den moeglichen verbilligten Sonderfertigkeiten oder
     *         {@code null} wenn es keine solche Liste gibt.
     */
    final VerbilligteSonderfertigkeit getVerbilligteSonderfertigkeitOffen() {

        final VerbilligteSonderfertigkeit verbilligung;
        verbilligung = verbilligteSonderfertigkeit;

        verbilligteSonderfertigkeit = null;

        return verbilligung;
    }

    /**
     * @return die Punkte die der Held durch Rasse, Kultur und Profession
     *         erhalten hat und auf Verbindungen verteilen darf.
     */
    public final int getVerbindungen() {
        return verbindungen;
    }

    /**
     * @return den naechsten Bonus fuer den es eine Auswahl zu treffen gilt,
     *         oder {@code null} wenn es keinen weiteren Bonus gibt.
     */
    final VorNachteilBonus getVorNachteilBonusOffen() {
        return vorNachteilBoniOffen.poll();
    }

    /**
     * @return die Zweitsprachen die die Kultur des Helden zur Auswahl bietet.
     *         Die Methode liefert nur einmal die Liste der Zweitsprachen. Bei
     *         folgenden Aufrufen bekommt man eine leere Liste falls keine
     *         weitere Zweitsprache mehr zu waehlen ist.
     */
    final List< Sprache > getZweitsprachen() {

        final List< Sprache > tmp = new ArrayList< Sprache >( zweitsprachen );

        zweitsprachen.clear();

        return tmp;
    }

    /**
     * Nach dem Aufruf von {@link #anwendenSonderfertigkeiten()} kann hier
     * abgefragt werden ob es der Benutzer noch eine Sonderfertigkeit fuer den
     * Helden aus einer Liste auswaehlen muss.
     *
     * @see #getSonderfertigkeitBonusOffen()
     * @return {@code true} wenn es noch einen offenen Sonderfertigkeit-Bonus
     *         gibt, sonst {@code false}.
     */
    final boolean hatSonderfertigkeitBonusOffen() {
        return !sonderfertigkeitBoniOffen.isEmpty();
    }

    /**
     * Nach dem Aufruf von {@link #anwendenTalentBoniLeittalente()} kann hier
     * abgefragt werden ob es einen Talent-Bonus vom
     * {@link TalentbonusTyp#ENTDECKER} gibt fuer den noch eine Auswahl
     * getroffen werden muss. Mehr zu diesem Typ Bonus siehe Regelwerk Seite 76,
     * Profession Entdecker, Talent-Boni Kategorie Wissen.
     *
     * @see #getTalentBonusEntdecker()
     * @return {@code true} wenn es einen Talent-Bonus vom Typ Entdecker gibt
     */
    final boolean hatTalentBonusEntdecker() {
        return null != talentBonusEntdecker;
    }

    /**
     * Nach dem Aufruf von {@link #anwendenTalentBoniLeittalente()} kann hier
     * abgefragt werden ob noch Talent-Boni existieren fuer die der Benutzer
     * eine Auswahl treffen muss. Diese Methode bezieht sich nur auf Talent-Boni
     * die der Held aus Rasse und Kultur erhalten hat.
     *
     * @see #getTalentBonusOffen()
     * @see #hatTalentBonusOffenProfession()
     * @return {@code true} wenn es noch einen offenen Talent-Bonus gibt, sonst
     *         {@code false}.
     */
    final boolean hatTalentBonusOffen() {
        return !talentBoniOffen.isEmpty();
    }

    /**
     * Nach dem Aufruf von {@link #anwendenTalentBoniLeittalente()} kann hier
     * abgefragt werden ob noch Talent-Boni existieren fuer die der Benutzer
     * eine Auswahl treffen muss. Diese Methode bezieht sich nur auf Talent-Boni
     * die der Held durch seine Profession erhalten hat.
     *
     * @see #getTalentBonusOffenProfession()
     * @see #hatTalentBonusOffen()
     * @return {@code true} wenn es noch einen offenen Talent-Bonus gibt, sonst
     *         {@code false}.
     */
    final boolean hatTalentBonusOffenProfession() {
        return !talentBoniOffenProfession.isEmpty();
    }

    /**
     * @return {@code true} wenn es noch einen Bonus gibt bei dem aus einer
     *         Liste eine Sonderfertigkeit gewaehlt werden muss die fuer den
     *         Helden dann als verbilligt gilt, sonst {@code false}.
     */
    final boolean hatVerbilligteSonderfertigkeitOffen() {
        return null != verbilligteSonderfertigkeit;
    }

    /**
     * @see #getVorNachteilBonusOffen()
     * @return {@code true}, wenn es noch einen Bonus gibt, bei dem der
     *         Benutzer einen Vor-/Nachteil auswaehlen muss, sonst {@code false}.
     */
    final boolean hatVorNachteilBonusOffen() {
        return !vorNachteilBoniOffen.isEmpty();
    }

    /**
     * @see #getAnzahlZusaetzlicheLeitzauber()
     * @return {@code true}, wenn der Held zusaetzliche Leitzauber waehlen
     *         darf, sonst {@code false}.
     */
    final boolean hatZusaetzlicheLeitzauber() {
        return ( 0 != anzahlZusaetzlicheLeitzauber );
    }

    /**
     * @return {@code true} wenn die Kultur mehrere Zweitsprachen zur Auswahl
     *         bietet, sonst {@code false}.
     */
    final boolean hatZweitsprachenAuswahl() {
        return zweitsprachen.size() > 1;
    }

    /**
     * Bestimmt ob eine Sonderfertigkeit mit AP oder GP bezahlt wird.
     *
     * @param wert
     *            fuer den die Zahlungsart festgelegt werden soll.
     * @param zahlungsart
     *            die angibt ob der Wert mit AP oder GP bezahlt wird.
     */
    public final void setZahlungsart( final SonderfertigkeitWert wert, final Zahlungsart zahlungsart ) {

        sonderfertigkeitenZahlungsart.put( wert, zahlungsart );

        setChanged();
        notifyObservers( wert.getSonderfertigkeit() );
    }

    /**
     * Die Zweitsprache des Helden.
     *
     * @param zweitsprache
     *            des Helden.
     */
    final void setZweitsprache( final Sprache zweitsprache ) {
        held.setZweitsprache( zweitsprache );
    }

    /**
     * Prueft ob eine Sonderfertigkeit bei der Erschaffung mit GP bezahlt wird.
     *
     * @param wert
     *            fuer den die Zahlungsart ermittelt werden soll.
     * @return {@code true} wenn die Sonderfertigkeit mit GP bezahlt wird,
     *         {@code false} wenn sie mit AP bezahlt wird.
     */
    public final boolean wirdMitGpBezahlt( final SonderfertigkeitWert wert ) {
        // Es gilt das Werte die noch nicht in der Map eingetragen sind, fuer
        // die get( wert ) noch null liefert, mit AP bezahlt werden.
        return sonderfertigkeitenZahlungsart.get( wert ) == Zahlungsart.GP;
    }
}
