/*
 * This file is part of the TDE Character Generator (TCG).
 *
 * TCG is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * TCG is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with TCG.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.bernhardjung.dsaprogramm.gui;

import static de.bernhardjung.dsaprogramm.i18n.I18n.i18n;

import java.awt.Color;
import java.util.ArrayList;
import java.util.List;
import java.util.Observable;
import java.util.Observer;

import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JSpinner;
import javax.swing.SwingConstants;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import com.jgoodies.forms.layout.CellConstraints;
import com.jgoodies.forms.layout.FormLayout;

import de.bernhardjung.dsaprogramm.basis.Held;
import de.bernhardjung.dsaprogramm.basis.KonfliktBeschreibung;
import de.bernhardjung.dsaprogramm.eigenschaften.Eigenschaft;
import de.bernhardjung.dsaprogramm.eigenschaften.EigenschaftWert;
import de.bernhardjung.dsaprogramm.util.Uebersetzbar;
import de.bernhardjung.dsaprogramm.util.Uebersetzer;
import de.bernhardjung.dsaprogramm.util.Util;

class PanelEigenschaften extends JPanel implements Observer, Uebersetzbar {

    /**
     * Die Ueberschrift fuer die Spalte die den Modifikator der Eigenschaft
     * anzeigt.
     */
    private static final String MOD = "PanelEigenschaften.Mod";

    /**
     * Die UEberschrift fuer die Spalte, die den Namen der Eigenschaft anzeigt.
     */
    private static final String EIGENSCHAFT = "PanelEigenschaften.Eigenschaft";

    /**
     * Die Ueberschrift fuer die Spalte die den Startwert der Eigenschaft 
     * anzeigt.
     */
    private static final String BASISSTUFE = "PanelEigenschaften.Basisstufe";

    /**
     * Die Ueberschrift fuer die Spalte die den aktuellen Wert des Helden
     * anzeigt.
     */
    private static final String AKTUELL = "PanelEigenschaften.Aktuell";

    /**
     * Der Text mit dem angegeben wird wie viele Punkte schon auf die
     * Eigenschaften verteilt worden sind.
     */
    private static final String VERTEILT = "PanelEigenschaften.verteilt";

    /**
     * Die Label die den Namen der Eigenschaft anzeigen.
     * Die Label in der Liste sind den Eigenschaften ueber ihren Index
     * zugeordnet d.h. an Position i steht das Label fuer die Eigenschaft mit
     * index i.
     */
    private final List< JLabel > eigenschaftenLabel;

    /**
     * Die Label die die Gesamtstufe der Eigenschaften anzeigen.
     * Die Label in der Liste sind den Eigenschaften ueber ihren Index
     * zugeordnet d.h. an Position i steht das Label fuer die Eigenschaft mit
     * index i.
     */
    private final List< JLabel > gesamtstufenLabel;

    /**
     * Der Held, dessen Werte angezeigt werden. 
     */
    private Held held;
    
    /**
     * Die Label die die Modifikation der Eigenschaften anzeigen.
     * Die Label in der Liste sind den Eigenschaften ueber ihren Index
     * zugeordnet d.h. an Position i steht das Label fuer die Eigenschaft mit
     * index i.
     */
    private final List< JLabel > modifikationenLabel;

    /**
     * Das Label mit dem angezeigt wird wie viel Punkte schon auf die
     * Eigenschaften verteilt worden sind.
     */
    private final JLabel verteiltLabel;

    /**
     * Die Spinner mit denen die Startstufe der Eigenschaft festgelegt wird.
     * Die Spinner in der Liste sind den Eigenschaften ueber ihren Index
     * zugeordnet d.h. an Position i steht der Spinner fuer die Eigenschaft mit
     * index i.
     */
    private final List< JSpinner > basisstufenSpinner;

    private final Uebersetzer uebersetzer;
    
    /**
     * Die Objekte die den Wert aus den Spinnern mit den Werten des Helden
     * synchronisieren.
     * Die Synchronisierer in der Liste sind den Eigenschaften/Spinnern ueber
     * ihren Index zugeordnet d.h. an Position i steht der Synchronisierer fuer
     * die Eigenschaft mit index i.
     */
    private final List< WertSynchronisierer > wertSynchronisierer;

    /** Erzeugt das Panel und ordnet die Komponenten darauf an. */
    PanelEigenschaften() {

        uebersetzer = new Uebersetzer();
        
        eigenschaftenLabel = new ArrayList< JLabel >( 9 );

        modifikationenLabel = new ArrayList< JLabel >( 9 );

        gesamtstufenLabel = new ArrayList< JLabel >( 9 );

        verteiltLabel = new JLabel( i18n( VERTEILT ) );

        verteiltLabel.setHorizontalAlignment( SwingConstants.CENTER );

        basisstufenSpinner = new ArrayList< JSpinner >( 9 );

        wertSynchronisierer = new ArrayList< WertSynchronisierer >( 9 );

        final JLabel ueberschriftAktuell = uebersetzer.erzeugeUeberschrift(
            AKTUELL );
        
        final JLabel ueberschriftBasisstufe = uebersetzer.erzeugeUeberschrift(
            BASISSTUFE );

        final JLabel ueberschriftEigenschaft = uebersetzer.erzeugeUeberschrift(
            EIGENSCHAFT ); 
        
        final JLabel ueberschriftModifikation = uebersetzer.erzeugeUeberschrift(
            MOD );

        final Eigenschaft[] eigenschaften;
        eigenschaften = Eigenschaft.getEigenschaftenMitSo();

        for ( final Eigenschaft eigenschaft : eigenschaften ) {

            eigenschaftenLabel.add( uebersetzer.erzeugeLabel( 
                    eigenschaft.getLokalisierungsschluesselName() ) );

            modifikationenLabel.add( Util.erzeugeLabelRechtsbuendig() );

            gesamtstufenLabel.add( Util.erzeugeLabelRechtsbuendig() );

            final JSpinner spinner = Util.erzeugeSpinner( 0, 0, 99, 1 );
            basisstufenSpinner.add( spinner );

            final WertSynchronisierer synchronisierer =
                new WertSynchronisierer( spinner );

            spinner.addChangeListener( synchronisierer );

            wertSynchronisierer.add( synchronisierer );

        }

        final FormLayout layout = new FormLayout(
                //Eigensch.    Mod.     +    Spinner    =      Wert
                "7dlu, P:G, 4dlu, P:G, 4dlu, P, 4dlu, P:G, 4dlu, P, 2dlu, P:G, 7dlu",
                "7dlu, P, 4dlu, P, 4dlu, P, 4dlu, P, 4dlu, P, 4dlu, P, " +
                "4dlu, P, 4dlu, P, 4dlu, P, 4dlu, P, 4dlu, P, 7dlu");

        setLayout( layout );

        final CellConstraints cc = new CellConstraints();

        add( verteiltLabel, cc.xyw( 2, 2, 11 ) );

        // Ueberschriften der Spalten
        add( ueberschriftEigenschaft,  cc.xy(  2, 4 ) );
        add( ueberschriftModifikation, cc.xy(  4, 4 ) );
        add( ueberschriftBasisstufe,   cc.xy(  8, 4 ) );
        add( ueberschriftAktuell,      cc.xy( 12, 4 ) );

        // Die restlichen Zeilen aufbauen
        for ( int i = 0 ; i < eigenschaften.length ; ++i ) {

            final int y = 2 * i + 6;
            int x = 0;

            add( eigenschaftenLabel.get( i ), cc.xy( x+=2, y ) );
            add( modifikationenLabel.get( i ),     cc.xy( x+=2, y ) );
            add( new JLabel( "+" ),           cc.xy( x+=2, y ) );
            add( basisstufenSpinner.get( i ),        cc.xy( x+=2, y ) );
            add( new JLabel( "=" ),           cc.xy( x+=2, y ) );
            add( gesamtstufenLabel.get( i ),       cc.xy( x+=2, y ) );
        }
    }

    public void neuUebersetzen() {

        final int verteiltePunkte = held.getVerteilteEigenschaftspunkte();

        final int verteilbar = Eigenschaft.getMaxVerteilbar( held );

        verteiltLabel.setText( String.format( 
                i18n( VERTEILT ),
                verteiltePunkte,
                verteilbar ) );
        
        uebersetzer.neuUebersetzen();
    }
    
    /**
     * Bestimmt den Helden der gerade erschaffen wird.
     *
     * @param held der gerade erschaffen wird.
     */
    public final void setHeld( final Held held ) {

        this.held = held;
        
        for ( final Eigenschaft eigenschaft :
            Eigenschaft.getEigenschaftenMitSo() ){

            final int index = eigenschaft.getIndex();

            wertSynchronisierer.get( index ).setWert(
                    held.getWert( eigenschaft ) );
        }

        held.addObserver( this );

        update( held, null );
    }

    /**
     * Aktualisiert die angezeigten Werte.
     * Veraenderte Modifikationen werden uebernommen, ebenso die veraenderte
     * Gesamstufe der Eigenschaften.
     * Regelkonflikte werden angezeigt.
     *
     * @param o der ueberwachte Held.
     * @param arg ungenutzt.
     */
    public void update( final Observable o, final Object arg ) {

        final Held held = (Held) o;

        final int verteiltePunkte = held.getVerteilteEigenschaftspunkte();

        final int verteilbar = Eigenschaft.getMaxVerteilbar( held );

        verteiltLabel.setText( String.format( 
                i18n( VERTEILT ),
                verteiltePunkte,
                verteilbar ) );

        if ( verteilbar < verteiltePunkte ) {

            verteiltLabel.setForeground( Util.FEHLERFARBE );

        } else {

            verteiltLabel.setForeground( getForeground() );

        }

        for ( final Eigenschaft eigenschaft :
            Eigenschaft.getEigenschaftenMitSo() ){

            final int index = eigenschaft.getIndex();

            final EigenschaftWert wert = held.getWert( eigenschaft );

            final KonfliktBeschreibung beschreibung = wert.pruefen();

            if( beschreibung.enthaeltKonflikt() ) {

                eigenschaftenLabel.get( index ).setToolTipText(
                        beschreibung.toHtml() );

                eigenschaftenLabel.get( index ).setForeground( Color.RED );

            } else {

                eigenschaftenLabel.get( index ).setToolTipText( null );

                eigenschaftenLabel.get( index ).setForeground( getForeground() );

            }

            modifikationenLabel.get( index ).setText(
                    String.valueOf( wert.getModifikation() ) );


            final JSpinner spinner = basisstufenSpinner.get( index );

            spinner.setValue( held.getWert( eigenschaft ).getBasisstufe() );


            gesamtstufenLabel.get( index ).setText(
                    String.valueOf( wert.getStufe() ) );

        }
    }

    /**
     * Synchronisiert einen JSpinner und einen {@link EigenschaftWert}.
     * Lauscht auf Aenderungen am Spinner und uebertraegt den Wert aus
     * dem Spinner als Startstufe in den {@link EigenschaftWert}.
     * */
    private class WertSynchronisierer implements ChangeListener {

        /** Der Spinner mit dem der EigenschaftWert synchronisiert wird. */
        private final JSpinner spinner;

        /** In den der Wert aus dem Spinner eingetragen wird. */
        private EigenschaftWert wert;

        /**
         * @param derSpinner dessen Wert ausgelesen wird.
         */
        private WertSynchronisierer( final JSpinner derSpinner ) {
            spinner = derSpinner;
        }

        /**
         * Uebertraegt den Wert aus dem Spinner in den EigenschaftWert.
         * @param e ungenutzt
         */
        public void stateChanged( final ChangeEvent e ) {

            if ( null != wert ) {

                wert.setBasisstufe( (Integer) spinner.getValue() );
            }
        }

        /**
         * Aendert den Wert in den der Wert aus dem Spinner eingetragen wird.
         * Setzt den Spinner auf die Startstufe aus dem EigenschaftWert
         *
         * @param neuerWert der aktualisert werden soll.
         */
        public void setWert( final EigenschaftWert neuerWert ) {

            wert = neuerWert;

            spinner.setValue( wert.getBasisstufe() );
        }
    }
}
