/*
 * This file is part of the TDE Character Generator (TCG).
 *
 * TCG is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * TCG is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with TCG.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.bernhardjung.dsaprogramm.gui.bonus;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.swing.AbstractCellEditor;
import javax.swing.JLabel;
import javax.swing.JRadioButton;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.SwingConstants;
import javax.swing.table.AbstractTableModel;
import javax.swing.table.DefaultTableCellRenderer;
import javax.swing.table.TableCellEditor;
import javax.swing.table.TableCellRenderer;
import javax.swing.table.TableColumn;

import de.bernhardjung.dsaprogramm.basis.Held;
import de.bernhardjung.dsaprogramm.bausteine.bonus.TalentbonusFest;
import de.bernhardjung.dsaprogramm.i18n.I18n;
import de.bernhardjung.dsaprogramm.talente.Talent;
import de.bernhardjung.dsaprogramm.talente.TalentWert;
import de.bernhardjung.dsaprogramm.util.Uebersetzbar;
import de.bernhardjung.dsaprogramm.util.Util;

public class PanelTalentBonusFestEntdecker
extends PanelTalentBonus 
implements Uebersetzbar {

    /**
     * Die Anzahl Spalten die vor der ersten Spalte mit den Radiobuttons
     * steht. Es sind 3 Spalten. Eine mit den Talente, eine mit der
     * Stufe der Talente und eine mit dem Stufenboni durch die Profession.
     */
    private static final int OFFSET = 3;

    /**
     * Die Hintergrundfarbe fuer die Tabelle. Wird benoetigt, weil sonst unter
     * GTK-Look & Feel alles auf weiss umgestellt wird.
     */
    private final Color hintergrund;

    /** Der Held fuer den die Auswahl getroffen wird. */
    private Held held;

    /** Speichert den frei gewordenen Bonus eines Talent. */
    private final Map< Talent, Integer > freiGewordeneStufenboniMap =
        new HashMap< Talent, Integer >();

    /**
     * Die Stufenboni die frei geworden sind, weil einer der
     * {@link #stufenboni} auf ein Talent verteilt wurde das schon einen
     * Bonus durch die Profession erhalten hat.
     */
    private final List< Integer > freiGewordeneStufenboni =
        new ArrayList< Integer >();

    /**
     * Speichert die ausgewaehlten Talente.
     * Der gespeicherte Integer Wert gibt den Index eines Talents in
     * {@link #talente} an. Anderer Erklaerung: Die Zahl gibt an in
     * welcher Zeile der Tabelle ein Knopf aktiviert ist.
     * An Position 0 in der Liste steht das Talent auf das der Bonus
     * an Position 0 in {@link #stufenboni} verteilt wurde.
     */
    private final List< Integer > gewaehlteTalente =
        new ArrayList< Integer >();

    /**
     * Das Tabellenmodell fuer den Dialog. Der ganze Dialog ist als Tabelle
     * implementiert. Auf diese Weise konnte ich das dynamische einfuegen
     * und entfernen von Spalten (fuer mich) am einfachsten umsetzen.
     */
    private final EntdeckerTableModel modell;

    /**
     * Die Stufenboni, die es durch den Talent-Bonus zu verteilen gibt.
     * Das sind die +5, +4 und +2.
     */
    private final List< Integer > stufenboni = new ArrayList< Integer >();

    /** Die Tabelle die die ganze Benutzerschnittstelle bildet. */
    private final JTable tabelle;

    /** Die Talent auf die die Boni verteilt werden sollen. */
    private final List< Talent > talente = new ArrayList< Talent >();

    public PanelTalentBonusFestEntdecker() {

        modell = new EntdeckerTableModel();

        // Es muss eine Kopie sein, sont wird es nicht angenommen.
        hintergrund = new Color( this.getBackground().getRGB() );

        tabelle = new JTable( modell );
        tabelle.setRowMargin( 0 );
        final DefaultTableCellRenderer renderer =
            (DefaultTableCellRenderer)tabelle.getDefaultRenderer( Object.class );
        renderer.setHorizontalAlignment( SwingConstants.RIGHT );
        renderer.setOpaque( true );
        renderer.setBackground( hintergrund );
        final DefaultTableCellRenderer zentriertRenderer =
            new DefaultTableCellRenderer();
        zentriertRenderer.setHorizontalAlignment( SwingConstants.CENTER );
        zentriertRenderer.setOpaque( true );
        zentriertRenderer.setBackground( hintergrund );
        tabelle.setDefaultRenderer( String.class, zentriertRenderer );
        tabelle.setDefaultRenderer( Integer.class, new RadioButtonCellRenderer() );
        tabelle.setDefaultEditor  ( Integer.class, new RadioButtonCellEditor  () );
        tabelle.getTableHeader().setReorderingAllowed( false );
        tabelle.setCellSelectionEnabled( false );
        tabelle.setShowGrid( false );
        final int fontSize = getFont().getSize();
        tabelle.setRowHeight( fontSize * 2 );
        tabelle.setFocusable( false );
        tabelle.setAutoResizeMode( JTable.AUTO_RESIZE_LAST_COLUMN );
        tabelle.setBackground( getBackground() );

        setLayout( new BorderLayout() );

        final JScrollPane scrollPane = Util.erzeugeScrollPane( tabelle );

        add( scrollPane );
    }

    /** {@inheritDoc} */
    @Override
    public final List<TalentWert> getTalentBoni() {

        final List< TalentWert > talentBoni = new ArrayList< TalentWert >();

        for ( int i = 0, n = gewaehlteTalente.size() ; i < n ; ++i ) {

            final Talent talent = talente.get( gewaehlteTalente.get( i ) );

            final int bonus;
            if ( i < stufenboni.size() ) {

                bonus = stufenboni.get( i );

            } else {

                bonus = freiGewordeneStufenboni.get( i - stufenboni.size() );

            }

            final TalentWert wert = new TalentWert( talent, bonus );

            talentBoni.add( wert );
        }

        return talentBoni;
    }

    public void neuUebersetzen() {
        
        modell.neuUebersetzen();
    }
    
    /**
     * Bestimmt den Held und den Bonus fuer die eine Auswahl getroffen werden
     * soll.
     *
     * @param held fuer den der Bonus sein soll.
     * @param bonus der die Auswahl anbietet.
     */
    public final void setBonus( final Held held, final TalentbonusFest bonus ) {

        modell.setBonus( held, bonus );
    }

    /**
     * Setzt die Breite einer Tabellenspalte auf einen festen Wert.
     * Die Breite Spalte ist nicht mehr aenderbar.
     *
     * @param spaltenIndex Index der Spalte deren Breite festgelegt werden
     *      soll.
     * @param breite der Spalte in Pixel.
     */
    public void setFesteSpaltenbreite( final int spaltenIndex, final int breite ) {
        final TableColumn spalte;
        spalte = tabelle.getColumnModel().getColumn( spaltenIndex );
        spalte.setMinWidth( breite );
        spalte.setMaxWidth( breite );
        spalte.setResizable( false );
    }

    /**
     * Macht die Breite eine Spalte aenderbar, setzt die MinWidth auf 0
     * und die MaxWidth auf Integer.MAX_VALUE.
     *
     * @param spaltenIndex Index der Spalte.
     */
    public void setSpalteVergroesserbar( final int spaltenIndex ) {
        final TableColumn spalte;
        spalte = tabelle.getColumnModel().getColumn( spaltenIndex );
        spalte.setMinWidth( 0 );
        spalte.setMaxWidth( Integer.MAX_VALUE );
        spalte.setResizable( true );
    }

    /**
     * Das Tabellenmodell das fuer das ganze Aussehn und Verhalten der
     * Benutzeroberflaeche zustaendig ist.
     */
    private class EntdeckerTableModel 
    extends AbstractTableModel
    implements Uebersetzbar {

        /**
         * Die Breite der Tabellenspalte in der die Namen der Talente stehen.
         */
        private static final int BREITE_SPALTE_TALENTE = 140;

        /**
         * Die Breite der Tabellenspalte in der die aktuelle Stufe, ohne den
         * Professionsbonus, steht, die der Held in dem Talent hat.
         */
        private static final int BREITE_SPALTE_AKTUELL = 60;

        /**
         * Die Breite der Tabellenspalte in der der Bonus steht, den der Held
         * durch die Profession auf das Talent erhalten hat.
         */
        private static final int BREITE_SPALTE_PROF = 40;

        /** Die Spalte der Tabelle in der die Talente stehen. */
        private static final int SPALTE_TALENTE = 0;

        /**
         * Die Spalte der Tabelle in der die aktuellen Stufen der Talente.
         */
        private static final int SPALTE_AKTUELL = 1;

        /**
         * Die Spalte der Tabelle in der der Stufenbonus durch die Profession
         * steht.
         */
        private static final int SPALTE_PROFESSION = 2;

        /** Die letzte Spalte der Tabelle, die leer ist. */
        private int spalteLeer;

        /**
         * Wird aufgerufen wenn einer der Auswahlknoepfe deaktiviert wird.
         *
         * @param zeile der Tabelle in der der Auswahlknopf steht.
         * @param spalte der Tabelle in der der Auswahlknopf steht.
         */
        private void abgewaehlt( final int zeile, final int spalte ) {

            // Ein Auswahlknopf wurde deaktiviert
            // => Ein Talent wurde abgewaehlt.
            // Wurde durch das Anwaehlen dieses Talents vorher ein
            // Professions-Bonus frei, so muss dies wieder rueckgaengig gemacht
            // werden.

            final Talent talent = talente.get( zeile );

            if ( ( spalte - OFFSET ) < stufenboni.size()
              && freiGewordeneStufenboniMap.containsKey( talent ) ) {
                // Im vorderen Bereich wurde ein Talent abgewaehlt.
                // Die Map enthaelt dieses Talent
                //  => beim anwaehlen wurde ein Bonus frei, der jetzt wieder
                //     gesetzt werden muss

                // Bonus heraussuchen
                final int bonus = freiGewordeneStufenboniMap.get( talent );

                // Der Bonus steht nicht laenger zur Auswahl.
                // Er muss aus dem hinteren Bereich entfernt werden.
                for ( int i = 0, n = freiGewordeneStufenboni.size()
                        ; i < n
                        ; ++i ) {

                    if ( freiGewordeneStufenboni.get( i ) == bonus ) {

                        freiGewordeneStufenboni.remove( i );

                        gewaehlteTalente.remove( stufenboni.size() + i );

                        --spalteLeer;

                        break;
                    }
                }

                held.einfuegenProfession( new TalentWert( talent, bonus ) );
            }
        }

        /**
         * Wird aufgerufen bevor einer der Auswahlknoepfe aktiviert wird.
         *
         * @param zeile der Tabelle in der der Knopf steht.
         * @param spalte der Tabelle in der der Knopf steht.
         */
        private void angewaehlt( final int zeile, final int spalte ) {

            // Auf folgendes muss geachtet werden.
            // Wird ein Bonus auf ein Talent verteilt das schon einen
            // Bonus durch die Profession erhalten hat, muss der schon
            // erhaltene Bonus frei gesetzt werden.
            final Talent talent = talente.get( zeile );

            if ( ( spalte - OFFSET ) < stufenboni.size()
              && held.hatVorgegebenProf( talent ) ) {
                // Es wurde im vorderen Bereich ein Bonus auf ein Talent
                // gesetzt das durch die Profession schon
                // einen Bonus erhalten hat.
                // => der schon erhaltene Bonus wird frei.

                // Dieser Bonus wird frei.
                final int bonus =
                    held.getVorgegebenProf( talent ).getUnmodifizierteStufe();

                // Der Held hat ihn jetzt nicht mehr.
                held.entfernenProfession( talent );

                // Den Bonus hinten in die Tabelle einfuegen.
                freiGewordeneStufenboni.add( bonus );

                // Eine Zeile in der Spalte suchen auf die der Bonus
                // verteilt werden kann.
                for( int z = 0, n = getRowCount() ; z < n ; ++z ) {

                    final Talent t = talente.get( z );
                    // Die Position muss frei sein, d.h in keiner anderen
                    // Spalte der Zeile darf ein Knopf aktiv sein und
                    // es darf auch nicht die aktuelle Zeile zeile sein,
                    // da der Knopf hier gleich aktiviert wird.
                    if ( ( z != zeile && positionIstFrei( z ) )
                            // Die in der aktuellen Spalte gewaehlte Zeile wird
                            // gleich auf zeile geaendert, daher kann man sie
                            // als frei nehmen, aber nur wenn ...
                            || ( z == gewaehlteTalente.get( spalte - OFFSET )
                                    // ... es hier kein Bonus freigesetzt wurde.
                                    && ! freiGewordeneStufenboniMap.containsKey( t ) ) ) {
                        // Eine freie Position gefunden.
                        gewaehlteTalente.add( z );
                        break;
                    }
                }
                ++spalteLeer;

                freiGewordeneStufenboniMap.put( talent, bonus );
            }
        }

        /**
         * @return Object.class fuer die Spalte mit den Talenten und die leere
         *      Spalte, String.class fuer die Spalte mit den Talentstufen/-boni
         *      und Integer.class fuer die Spalten mit den Auswahlknoepfen.
         */
        @Override
        public Class<?> getColumnClass( final int spalte ) {

            if( ( spalte == SPALTE_TALENTE )
             || ( spalte == spalteLeer ) ) {

                return Object.class;

            } else if( ( spalte == SPALTE_AKTUELL )
                    || ( spalte == SPALTE_PROFESSION ) ) {

               return String.class;

            } else {

                return Integer.class;

            }
        }

        /** {@inheritDoc} */
        public int getColumnCount() {
            return OFFSET // Drei Spalten vorne,
                    + stufenboni.size()  // fuer jeden Bonus eine Spalte,
                    // fuer jeden frei gewordenen Bonus eine Spalte
                    + freiGewordeneStufenboni.size()
                    + 1; // und eine leere Spalte ganz hinten in der Tabelle.
        }

        @Override
        public String getColumnName( final int spalte ) {

            // Vorne sind die Spalten mit den stufenboni die durch den
            // Talent-Bonus in jeden Fall vergeben werden.
            final int vorne = ( OFFSET + stufenboni.size() );

            // Hinten sind die Spalten mit den frei gewordenen Stufenboni.
            final int hinten = vorne + freiGewordeneStufenboni.size();

            if ( SPALTE_TALENTE == spalte ) {

                // Die Spalte mit den Talenten hat keine Ueberschrift.
                // Es muss ein Leerzeichen sein, sonst die Hoehe der Zeile mit
                // den Spaltenueberschriften 0.
                return " ";

            } else if ( SPALTE_AKTUELL == spalte ) {

                return I18n.i18n( "PanelTalentBonusFest.aktuell" );

            } else if ( SPALTE_PROFESSION == spalte ) {

                return I18n.i18n( "PanelTalentBonusFestEntdecker.Profession" );

            } else if ( vorne > spalte ) {

                final int index = ( spalte - OFFSET );

                return String.format( "%1$+d", stufenboni.get( index ) );

            } else if ( hinten > spalte ){

                final int index = ( spalte - vorne );

                final int bonus = freiGewordeneStufenboni.get( index );

                return String.format( "%1$+d", bonus );

            } else {
                // Die leere Spalte
                return "";
            }


        }

        /**
         * @return die Anzahl der Talente.
         */
        public final int getRowCount() {
            return talente.size();
        }

        public final Object getValueAt( final int zeile, final int spalte ) {

            final Talent talent = talente.get( zeile );

            if ( SPALTE_TALENTE == spalte ) {

                return talent.getName();

            } else if( SPALTE_AKTUELL == spalte ) {

                if ( held.hat( talent ) ) {
                    // In der Spalte soll die Stufe des Talents ohne den
                    // Bonus aus der Profession angezeigt werden.
                    int stufe = held.getStufe( talent );

                    if ( held.hatVorgegebenProf( talent ) ) {

                        final TalentWert wert = held.getVorgegebenProf( talent);

                        stufe -= wert.getUnmodifizierteStufe();

                    }

                    return stufe;

                } else {

                    // Der Held besitzt das Talent noch nicht.
                    return "-";

                }

            } else if ( SPALTE_PROFESSION == spalte ) {

                if ( held.hatVorgegebenProf( talent ) ) {

                    final TalentWert wert = held.getVorgegebenProf( talent );

                    final int stufe = wert.getUnmodifizierteStufe();

                    return String.format( "%1$+d", stufe );

                } else {

                    return "-";

                }

            } else if ( spalte == spalteLeer) {

                return "";

            } else {

                return gewaehlteTalente.get( spalte - OFFSET );

            }
        }

        @Override
        public boolean isCellEditable( final int zeile, final int spalte ) {

            final int bonus = ( spalte - OFFSET );

            final Talent talent = talente.get( zeile );

            return ( spalte >= OFFSET ) && ( spalte != spalteLeer )
                && ( bonus < stufenboni.size()
                    || ( ! freiGewordeneStufenboniMap.containsKey( talent )
                            && ! held.hatVorgegebenProf( talent ) )
                    || ( freiGewordeneStufenboniMap.get( talent )
                            == freiGewordeneStufenboni.get( bonus - stufenboni.size() ) ) );

        }

        public void neuUebersetzen() {
            
            final List< Talent > talenteListe = new ArrayList< Talent >();
            
            for ( final int index : gewaehlteTalente ) {
                
                talenteListe.add( talente.get( index ) );
            }
            
            Collections.sort( talente, Talent.NAME_COMPARATOR );
            
            for ( int i = 0, n = gewaehlteTalente.size() ; i < n ; ++i ) {
                
                final int index = talente.indexOf( talenteListe.get( i ) );
                
                gewaehlteTalente.set( i, index );
            }
            
            fireTableDataChanged();
        }
        
        /**
         * Prueft ob ein Talent/eine Zeile angewaehlt werden kann.
         * Eine Talent gilt nur als frei, wenn es noch nicht angewaehlt ist
         * udn auf das Talent keinen Bonus durch die Profession vergeben wurde.
         *
         * @param zeile
         *            fuer die ueberpueft werden soll ob sie frei ist.
         * @return {@code true}, wenn in Zeile zeile noch kein Button
         *         angewaehlt ist, sonst {@code false}.
         */
        private boolean positionIstFrei( final int zeile ) {

            final Talent talent = talente.get( zeile );

            if ( held.hatVorgegebenProf( talent )
              || freiGewordeneStufenboniMap.containsKey( talent ) ) {

                return false;
            }

            // position sind die angewaehlten Zeilen.
            for ( final int position : gewaehlteTalente ) {

                if ( position == zeile ) {

                    return false;

                }
            }

            return true;
        }

        public void setBonus(
                final Held derHeld,
                final TalentbonusFest bonus ) {

            held = derHeld;

            freiGewordeneStufenboni.clear();
            freiGewordeneStufenboniMap.clear();

            stufenboni.clear();
            stufenboni.addAll( bonus.getBoni() );

            talente.clear();
            talente.addAll( bonus.getTalente( held ) );
            gewaehlteTalente.clear();
            for ( int i = 0, zeile = 0, n = stufenboni.size() ; i < n ; ++i ) {

                Talent talent = talente.get( zeile );
                while ( held.hatVorgegebenProf( talent ) ) {
                    ++zeile;
                    talent = talente.get( zeile );
                }

                gewaehlteTalente.add( zeile );

                zeile++;
            }

            spalteLeer = OFFSET + stufenboni.size();

            fireTableStructureChanged();

            setFesteSpaltenbreite( SPALTE_TALENTE, BREITE_SPALTE_TALENTE );
            setFesteSpaltenbreite( SPALTE_AKTUELL, BREITE_SPALTE_AKTUELL );
            setFesteSpaltenbreite( SPALTE_PROFESSION, BREITE_SPALTE_PROF );
            for( int i = 0, n = stufenboni.size() ; i < n ; ++i ) {

                setFesteSpaltenbreite( ( i + OFFSET ), 36 );
            }
            setSpalteVergroesserbar( spalteLeer );
        }

        @Override
        public void setValueAt( final Object value, final int zeile, final int spalte ) {

            // Es geht um den Bonus an Position bonus in stufenboni
            // bzw. um den Bonus an Position ( bonus - stufenboni.size() )
            // in freiGewordeneStufenboni.
            final int bonus = ( spalte - OFFSET );

            // Nur wenn eine Aenderung stattgefunden hat.
            if( gewaehlteTalente.get( bonus ) != zeile ) {

                final int alt = gewaehlteTalente.get( bonus );

                int tauschZeile = -1;
                int tauschSpalte = -1;

                // alle Spalte durchgehen und pruefen ob in der Zeile schon ein
                // Knopf angewaehlt ist.
                for ( int s = 0, n = gewaehlteTalente.size() ; s < n ; ++s ) {

                    if ( gewaehlteTalente.get( s ) == zeile ) {
                        // In Spalte s ist aktuell schon ein Knopf in zeile
                        // aktiviert.
                        tauschSpalte = ( s + OFFSET );
                        tauschZeile = alt;

                        break;
                    }
                }

                if ( tauschSpalte != -1 ) {
                    // Es wird getauscht

                    if ( bonus < stufenboni.size()
                      && ( tauschSpalte - OFFSET ) < stufenboni.size() ) {
                        // Tauschen im vorderen Bereich => keine Veraenderung

                        setValueAt( zeile, spalte );

                    } else if ( ( tauschSpalte - OFFSET ) < stufenboni.size() ) {
                        // Tausch zwischen vorderem und hinterem Bereich.
                        // geklickt in hinterem Bereich

                        angewaehlt( tauschZeile, tauschSpalte );
                        abgewaehlt( gewaehlteTalente.get( tauschSpalte-OFFSET ),
                                    tauschSpalte );

                        // Die hintere Spalte kann jetzt verschwunden sein, wir
                        // tun so, als haetten wir in die vordere geklickt.
                        setValueAt( tauschZeile, tauschSpalte );

                    } else if ( bonus < stufenboni.size() ){
                        // Tausch zwischen vorderm und hinterem Bereich.
                        // geklickt in vorderem Bereich.

                        angewaehlt( zeile, spalte );
                        abgewaehlt( gewaehlteTalente.get( bonus ), spalte );

                        setValueAt( zeile, spalte );

                    } else {
                        // Tausch im hinteren Bereich nichts zu tun.

                        setValueAt( zeile, spalte );
                    }

                } else {
                    // Es wird nicht getauscht

                    abgewaehlt( gewaehlteTalente.get( bonus ), spalte );
                    angewaehlt( zeile, spalte );

                    setValueAt( zeile, spalte );
                }

                fireTableStructureChanged();

                setFesteSpaltenbreite( SPALTE_TALENTE, BREITE_SPALTE_TALENTE );
                setFesteSpaltenbreite( SPALTE_AKTUELL, BREITE_SPALTE_AKTUELL );
                setFesteSpaltenbreite( SPALTE_PROFESSION, BREITE_SPALTE_PROF );
                for( int i = 0, n = stufenboni.size() ; i < n ; ++i ) {

                    setFesteSpaltenbreite( ( i + OFFSET ), 36 );
                }
                for( int i = 0, n = freiGewordeneStufenboni.size() ; i < n ; ++i ) {

                    setFesteSpaltenbreite(
                            ( i + OFFSET + stufenboni.size() ), 36 );
                }
                setSpalteVergroesserbar( spalteLeer );
            }

        }

        public void setValueAt( final int zeile, final int spalte ) {

            // alt ist die bisher angewaehlte Zeile.
            // diese wird nach der Methode frei sein.
            final int alt = gewaehlteTalente.get( spalte - OFFSET );

            // Die Liste der bisher gewaehlten Zeilen durchlaufen.
            for ( int i = 0, n = gewaehlteTalente.size() ; i < n ; ++i ) {

                if ( gewaehlteTalente.get( i ) == zeile ) {
                    // Die neu anzuwaehlende Zeile ist schon besetzt.

                    if ( isCellEditable( alt, i + OFFSET ) ) {
                        // wenn die Zelle frei ist, wird einfach getauscht.
                        gewaehlteTalente.set( i, alt );

                    } else {
                        // Es kann nicht einfach getauscht werden.
                        // Es werden einfach alle Zeilen von oben an durchsucht
                        // ob sie frei sind. j ist die Zeile.
                        for ( int j = 0, m = talente.size() ; j < m ; ++j ) {

                            if ( j != zeile
                                    && positionIstFrei( j ) ) {

                                gewaehlteTalente.set( i, j );
                                break;
                            }
                        }
                    }

                    break;
                }
            }

            gewaehlteTalente.set( ( spalte - OFFSET ), zeile );
        }
    }

    private class RadioButtonCellRenderer
    extends JRadioButton
    implements TableCellRenderer {

        private final JLabel label;

        public RadioButtonCellRenderer() {
            super();
            setHorizontalAlignment(JRadioButton.CENTER);
            setOpaque( true );
            label = new JLabel();
            label.setOpaque( true );
            label.setBackground( hintergrund );
        }

        public Component getTableCellRendererComponent(
                final JTable table,
                final Object value,
                final boolean isSelected,
                final boolean hasFocus,
                final int zeile,
                final int spalte ) {

            final int bonus = ( spalte - OFFSET );

            final Talent talent = talente.get( zeile );

            if ( ( bonus >= stufenboni.size() )
                    && ( held.hatVorgegebenProf( talent )
                            || ( freiGewordeneStufenboniMap.containsKey( talent ) )
                    && ( freiGewordeneStufenboniMap.get( talent )
                            != freiGewordeneStufenboni.get( bonus - stufenboni.size() ) ) ) ) {

                return label;

            }

            if( ( bonus >= stufenboni.size() )
                    && ( freiGewordeneStufenboniMap.containsKey( talent )
                            && ( freiGewordeneStufenboniMap.get( talent )
                            == freiGewordeneStufenboni.get( bonus - stufenboni.size() ) ) ) ) {

                setBackground( Color.GREEN );

            } else {

                setBackground( hintergrund );
            }

            setSelected( value.equals( zeile ) );

            return this;
        }

    }

    private static class RadioButtonCellEditor
    extends AbstractCellEditor
    implements TableCellEditor, ActionListener {

        private final JRadioButton button = new JRadioButton();

        public RadioButtonCellEditor() {
            button.setHorizontalAlignment(JRadioButton.CENTER);
            button.addActionListener(this);
            button.setOpaque(false);
        }

        public Component getTableCellEditorComponent(final JTable table,
                final Object value, final boolean isSelected, final int zeile, final int spalte) {

            button.setSelected(value.equals(zeile));

            return button;
        }

        public Object getCellEditorValue() {
            return null;
        }

        public void actionPerformed(final ActionEvent e) {
            stopCellEditing();
        }

    }

}
