/*
 * This file is part of the TDE Character Generator (TCG).
 *
 * TCG is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * TCG is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with TCG.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.bernhardjung.dsaprogramm.talente.kosten;

import de.bernhardjung.dsaprogramm.basis.Held;
import de.bernhardjung.dsaprogramm.basis.Lernmethode;
import de.bernhardjung.dsaprogramm.basis.SKT;
import de.bernhardjung.dsaprogramm.basis.varianten.Variante;
import de.bernhardjung.dsaprogramm.talente.Talent;
import de.bernhardjung.dsaprogramm.talente.Talentgruppe;
import de.bernhardjung.dsaprogramm.vornachteile.VorNachteil;
import de.bernhardjung.dsaprogramm.vornachteile.VorNachteilWert;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

public abstract class Kosten {

    /** Implementiert die Regeln die fuer die meisten Talente gueltig sind. */
    public static final Kosten TALENT = new KostenStandard();

    /** Implementiert die Regeln die fuer die Kampftechniken gelten. */
    public static final Kosten KAMPFTECHNIK = new KostenKampftechnik();

    /**
     * Implementiert die Regeln, die fuer die Talente gelten, die durch den
     * Vorteil Schlangenmensch verbilligt werden. Mit Ausnahme von Ringen.
     * Fuer Ringen kann auch {@link #KAMPFTECHNIK} verwendet werden.
     */
    public static final Kosten SCHLANGENMENSCH = new KostenSchlangenmensch();

    /**
     * Implementiert die Regeln die fuer Schriften und Wissenstalente gelten.
     */
    public static final Kosten SCHRIFT_WISSEN = new KostenSchriftWissen();

    /** Implementiert die Regeln die fuer die Sprachen gelten. */
    public static final Kosten SPRACHE = new KostenSprache();

    private VorNachteilWert unfaehigkeitTalent;

    private VorNachteilWert unfaehigkeitTalentgruppe;

    /**
     * @return eine Liste von Regelelementen von denen die Berechnung der Kosten
     *         abhaengig ist.
     */
    public Collection<Object> getAbhaengigkeiten() {

        final List< Object > abhaengigkeiten = new ArrayList< Object >( 2 );

        abhaengigkeiten.add( VorNachteil.UNFAEHIGKEIT_TALENT );

        abhaengigkeiten.add( VorNachteil.UNFAEHIGKEIT_TALENTGRUPPE );

        return abhaengigkeiten;
    }

    /**
     * Berechnet die Aktivierungskosten die ein Held fuer ein Talent bezahlen
     * muss.
     *
     * @param held
     *            der das Talent aktivieren will.
     * @param talent
     *            das der Held aktivieren will.
     * @param lernmethode
     *            die zum Aktivieren verwendet werden soll.
     * @return die AP, die {@code held} bezahlen muss, wenn er das Talent
     *         aktivieren will.
     */
    public int getAktivierungskosten(
            final Held held,
            final Talent talent,
            final Lernmethode lernmethode ){

        if ( lernmethode == Lernmethode.FREI )
            return 0;

        final int steigerungsfaktor = getSteigerungsspalte(
                held, talent, 0, lernmethode );

        if ( held.wirdErschaffen() ) {

            return SKT.getAktivierungskostenErschaffung( steigerungsfaktor );

        } else {

            return SKT.getAktivierungskosten( steigerungsfaktor );
        }
    }

    /**
     * Bestimmt den Steigerungsfaktor eines Talents fuer einen bestimmten Held.
     *
     * @param held fuer den der Steigerungsfaktor bestimmt werden soll.
     * @param talent dessen Steigernungsfaktor gesucht ist.
     * @param stufe Die aktuelle Stufe/Der aktuelle Talentwert des Talents.
     * @param lernmethode die zum Steigern verwendet werden soll.
     * @return den Steigerungsfaktor den das Talent fuer den Helden besitzt.
     */
    public int getSteigerungsspalte(
            final Held held,
            final Talent talent,
            final int stufe, 
            final Lernmethode lernmethode ) {

        int steigerungsfaktor = lernmethode.bestimmeSteigerungsfaktor( talent );

        setUnfaehigkeit( talent );

        if ( held.hat( unfaehigkeitTalent ) ) {

            ++steigerungsfaktor;
        }

        if ( held.hat( unfaehigkeitTalentgruppe ) ) {

            ++steigerungsfaktor;
        }

        return steigerungsfaktor;
    }


    /**
     * Berechnet die Abenteuerpunkte die ein Held bezahlen muss um ein Talent um
     * einen Punkt zu steigern.
     *
     * @param held der das Talent steigern will.
     * @param talent das gesteigert werden soll.
     * @param stufe Der aktuelle Talentwert/Die aktuelle Stufe des Talents.
     * @param lernmethode die zum Steigern verwendet werden soll.
     * @return Die Abenteuerpunkte, die {@code held} bezahlen muss, wenn er
     *      das Talent um einen Punkt steigern will.
     */
    public int getSteigerungskosten(
            final Held held,
            final Talent talent,
            final int stufe,
            final Lernmethode lernmethode ) {

        if ( lernmethode == Lernmethode.FREI )
            return 0;

        final int steigerungsfaktor =
            getSteigerungsspalte( held, talent, stufe, lernmethode );

        return SKT.getKosten( stufe+1, steigerungsfaktor );
    }

    /**
     * Stellt die Unfaehigkeiten auf das Talent ein und
     * initialisiert sie ggf.
     *
     * @param talent
     *            fuer die Unfaehgikeiten.
     */
    private void setUnfaehigkeit( final Talent talent ) {

        if ( null == unfaehigkeitTalent ) {

            unfaehigkeitTalent = new VorNachteilWert( 
                    VorNachteil.UNFAEHIGKEIT_TALENT,
                    new Variante( Talent.ABRICHTEN ) );

            unfaehigkeitTalentgruppe = new VorNachteilWert( 
                    VorNachteil.UNFAEHIGKEIT_TALENTGRUPPE,
                    new Variante( Talentgruppe.GESELLSCHAFT ) );
        }

        unfaehigkeitTalent.setVariante( talent );

        unfaehigkeitTalentgruppe.setVariante( talent.getTalentgruppe() );
    }
}
