/*
 * This file is part of the TDE Character Generator (TCG).
 *
 * TCG is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * TCG is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with TCG.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.bernhardjung.dsaprogramm.talente.voraussetzungen;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;

import de.bernhardjung.dsaprogramm.basis.Held;
import de.bernhardjung.dsaprogramm.basis.KonfliktBeschreibung;
import de.bernhardjung.dsaprogramm.i18n.I18n;
import de.bernhardjung.dsaprogramm.talente.Talent;
import de.bernhardjung.dsaprogramm.util.Strings;

class VoraussetzungSteigern extends Voraussetzung {
    
    /**
     * Die Regelelemente (Talente, Eigenschaften,...) von denen die
     * Voraussetzung abhaengt.
     */
    private final List< Object > abhaengigkeiten;
    
    /** Von denen der Held mindestens eines besitzen muss. */
    private final List< Talent > talente;
    
    /** Die Stufe die der Held in einem Talent mindestens erreichen muss. */
    private final int verlangteStufe;
    
    /**
     * 
     * @param dieStufe die in den Talenten/dem Talent vorausgesetzt wird.
     * @param dieTalente von denen der Held mindestens eines auf der 
     *      angegebenen Stufe besitzen muss.
     */
    VoraussetzungSteigern( final int dieStufe, final Talent ... dieTalente ) {
        
        talente = Arrays.asList( dieTalente );
        
        abhaengigkeiten = new ArrayList< Object >( talente );
        
        verlangteStufe = dieStufe;
    }
    
    /**
     * Macht die Voraussetzung, zusaetzlich zu den Talenten, von einem weiteren
     * Regelelement (z.B. einer Eigenschaft) abhaengig.
     * 
     * @param regelelement
     *            ein Vor-/Nachteil, eine Eigenschaft, ein Basiswert usw.
     */
    void ergaenzeAbhaengigkeit( final Object regelelement ) {
        abhaengigkeiten.add( regelelement );
    }
    
    /** @return die Talente. */
    public Collection<Object> getAbhaengigkeiten() {
        return abhaengigkeiten;
    }
    
    /** {@inheritDoc} */
    public KonfliktBeschreibung pruefen( 
            final Held held, 
            final Talent talent ) {

        final KonfliktBeschreibung beschreibung = new KonfliktBeschreibung();
    
        if ( held.hat( talent ) 
                && ( held.getStufe( talent ) > 10 ) ) {
            
            boolean erfuellt = false;
            
            for ( final Talent verlangtesTalent : talente ) {
                
                if ( held.hat( verlangtesTalent ) 
                && ( held.getStufe( verlangtesTalent ) >= verlangteStufe ) ) {
                    
                    erfuellt = true;
                    break;
                }
            }
            
            if ( ! erfuellt ) {
                
                beschreibung.einfuegen( String.format( 
                        I18n.i18n( "VoraussetzungSteigern" ),
                        Strings.erzeugeAufzaehlungMitOder( talente ),
                        verlangteStufe ) );
            }
        }
        
        return beschreibung;
    }

}
